class StringHelper:
    """
    Collection of helper methods for the string datatype.
    """

    @staticmethod
    def startswith_multi(text: str, prefixes: list) -> bool:
        """

        """
        for prefix in prefixes:
            if text.startswith(prefix):
                return True
        return False


    @staticmethod
    def remove_prefix(text: str, prefix: str) -> str:
        """
        Removes the specified prefix from the text.

        Parameters:
            text (str): Text to remove the prefix from.
            prefix (str): Prefix to remove from the text.

        Returns:
            str: The text minus the specified prefix.
        """
        if text.startswith(prefix):
            return text[len(prefix):]
        return text


    @staticmethod
    def remove_postfix(text, postfix):
        """
        Removes the specified postfix from the text.

        Parameters:
            text (str): Text to remove the postfix from.
            postfix (str): postfix to remove from the text.

        Returns:
            str: The text minus the specified postfix.
        """
        if text.endswith(postfix):
            return text[: - (len(postfix))]
        return text



import os
from pathlib import Path


class FileDetails:
    """

    """

    def __init__(self, path: Path, output_path: str, root: str):
        """

        """
        self.name = path.name
        self.name_without_suffix = path.name.replace(path.suffix, '')
        self.parent = str(path.parent)
        self.extension = path.suffix
        self.full_name = str(path)
        self.rel_parent = self.parent.replace(root, '')
        self.rel_full_name = self.full_name.replace(root, '')


import re

from .string_helper import StringHelper


class YAMLCommentHelper:
    """
    Collection of helper methods to be used with YAML comments.
    """


    @staticmethod
    def format_comment(comment: str) -> str:
        """
        Formats the comment by removing the YAML comment indicator '#',
        and removing newlines/blanks spaces from either sides of the string.

        Parameters:
            comment (str): The comment to format.

        Returns:
            str: Returns the formatted comment.
        """
        # Remove stray newlines.
        comment = StringHelper.remove_prefix(comment, '\n')
        comment = StringHelper.remove_postfix(comment, '\n')

        # Check we are not stripping comment indentation.
        if comment.strip().startswith('#'):
            comment = comment.strip()

        # Remove the comment identifier.
        comment = StringHelper.remove_prefix(comment, '# ')

        return comment


    @staticmethod
    def format_comment_lines(comment):
        """

        """
        lines = list()

        for line in comment.split('\n'):
            lines.append(YAMLCommentHelper.format_comment(line))

        return '\n'.join(lines)



import os
from operator import xor
from pathlib import Path
from typing import List

import click
from jinja2 import Template
from pyfiglet import Figlet
from ruamel.yaml.parser import ParserError

from .helpers.file_helper import FileHelper
from .models.file_details import FileDetails
from .wrappers.ruamel_yaml_wrapper import RuamelYAMLWrapper
from .yaml_docs_parser import YAMLDocsParser
from .wrappers.jinja_environment_wrapper import JinjaEnvironmentWrapper


@click.group()
@click.version_option("1.0.0")
def main():
    f = Figlet(font='slant', justify='')
    print('EMBEDDED...')
    print(f.renderText("YAML Docs"))
    pass


@main.command()
@click.option(
    '-p',
    '--path',
    'path',
    help='Root folder to search for YAML files in.',
    required=True,
    type=str
)
@click.option(
    '-a',
    '--pattern',
    'pattern',
    help='Search pattern to use when looking for YAML files.',
    required=True,
    type=str
)
@click.option(
    '-t',
    '--template-path',
    'template_path',
    help='Path to the Jinja2 template.',
    required=True,
    type=str
)
@click.option(
    '-o',
    '--output-path',
    'output_path',
    help='Folder path to output all the generated files to.',
    required=True,
    type=str
)
@click.option(
    '-e',
    '--exclude-comments',
    'exclude_comments',
    help='List of comment prefixes to exclude from the generated documentation.',
    multiple=True,
    required=False
)
@click.option(
    '-r',
    '--recurse',
    'recurse',
    help='Gets the items in the specified locations and in all child items of the locations. False by default.',
    is_flag=True,
    required=False
)
@click.option(
    '-w',
    '--overwrite',
    'overwrite',
    help='Overwrites any existing documents.',
    is_flag=True,
    required=False
)
@click.option(
    '-i',
    '--include-output-suffix',
    'include_output_suffix',
    help='Includes the files suffix for the output file.',
    is_flag=True,
    required=False
)
def generate(
    path: str,
    pattern: str,
    template_path: str,
    output_path: str,
    exclude_comments: list=None,
    recurse: bool = False,
    overwrite: bool = False,
    include_output_suffix: bool = False
):
    """

    """
    # Check if the paths are in the correct format.
    # if xor(os.path.isfile(path), os.path.isfile(output_path)):
    #     raise Exception('The "path" and "output_path" must either be both a file path or a directory path.')

    # Prepare the libaries.
    yaml_parser = RuamelYAMLWrapper()
    docs_parser = YAMLDocsParser(yaml_parser)

    # Do we want to recursively search?
    files = FileHelper.get_files(path, pattern, recurse)
    print(f'Found {len(files)} YAML files to generate the docs for...\n')

    # If there are no files to process throw an error.
    if len(files) == 0:
        raise Exception('No files were found to generate docs for. Double check your search pattern.')

    # Loop through each of the files and generate the docs.
    for idx, file in enumerate(files):
        try:
            # Get the file details.
            fd = FileDetails(file, output_path, path)
            print(f'[{idx + 1}/{len(files)}] - {fd.rel_full_name}')

            # Load the YAML file.
            try:
                yaml = yaml_parser.load_from_file(fd.full_name)
            except ParserError as e:
                print('There was an error in the YAML file.')
                raise

            if not isinstance(exclude_comments, list):
                exclude_comments = [exclude_comments]

            # Update the YAML file with docs using ByRef.
            print('Generating the docs...')
            docs_parser.extract_docs(yaml, exclude_comments)

            # Render the doc file using the template.
            try:
                environment = JinjaEnvironmentWrapper()
                template = environment.from_file(template_path)
                result = template.render(
                    _file=fd,
                    _yaml=yaml
                )
            except Exception as e:
                print('There was an error with the supplied jinja2 template...')
                raise

            # Output the generated doc.
            try:
                print(f'Outputting the docs...')

                if include_output_suffix:
                    output_file=FileHelper.join_paths(
                        output_path,
                        fd.rel_parent,
                        fd.name + '.md'
                    )
                else:
                    output_file=FileHelper.join_paths(
                        output_path,
                        fd.rel_parent,
                        fd.name_without_suffix + '.md'
                    )

                mode = 'x'
                if overwrite:
                    mode = 'w'

                with FileHelper.open_makedirs(output_file, mode) as f:
                    f.write(result)

            except Exception as e:
                print('There was an error outputting the doc.')
                raise

        except Exception as e:
            print(f'Error at file "{fd.full_name}"')
            print(e)

        print('\n')


if __name__ == '__main__':
    main()



from ruamel.yaml import YAML, safe_load
from ruamel.yaml.compat import StringIO


class RuamelYAMLWrapper(YAML):
    """
    Custom implementation of the ruamel library.
    """


    def dump_str(self, data: any, **kwargs) -> str:
        """
        Return the YAML object as a string.

        Properties:
            data<any>: The YAML object.
            **kwargs: Options for the YAML dump.

        Returns:
            The YAML object as a string.
        """
        stream = StringIO()

        super().dump(data, stream, **kwargs)

        output_str = stream.getvalue()

        stream.close()

        return output_str


    def load_from_file(self, file: str) -> any:
        """
        Loads a YAML document from a file, includes any metadata, comments etc...

        Parameters:
            file (str): The file path to load the YAML from.

        Returns:
            The parsed YAML document.
        """
        return super().load(open(file).read())


    def safe_load(self, stream) -> any:
        """
        Safely loads the YAML document, excluding any metadata.

        Parameters:
            stream (any): The YAML stream to parse.

        Returns:
            The parsed YAML document.
        """
        return safe_load(stream)



class JinjaFunctions():
    """

    """


    @staticmethod
    def postindent_spaces(s: str, indent: int) -> str:
        """
        Indents the specified string with X amount of spaces, at the end.

        Parameters:
            s (str): The string to post indent with spaces.
            indent (int): Number of spaces to indent the string with.

        Returns:
            The indented string.
        """
        return f"{{:<{indent}}}".format(s)


    @staticmethod
    def preindent_spaces(s: str, indent: int) -> str:
        """
        Indents the specified string with X amount of spaces, at the start.

        Parameters:
            s (str): The string to pre indent with spaces.
            indent (int): Number of spaces to indent the string with.

        Returns:
            The indented string.
        """
        return f"{{:>{indent}}}".format(s)


    @staticmethod
    def align_multiline(block: str, indent: int, newline: bool):
        """
        Aligns a multi-line string by adding the relevant newlines and indentation.

        Parameters:
            block (str): The string block to indent.
            indent (int): Number of spaces to indent the string block by.
            newline (bool): True or False if the string block should start with a newline.

        Returns:
            A block string that has the lines aligned correctly.
        """
        # Variable to save the result.
        result = ''

        # Add a newline to the start of the string.
        if newline:
            result += '\n'

        # Indent each of the lines to match.
        for line in block.split('\n'):
            result += JinjaFunctions.preindent_spaces(line, indent) + '\n'

        return result


from jinja2.filters import prepare_select_or_reject, contextfilter


class JinjaFilters:
    """
    Collection of custom Jinja Filters.
    """


    @staticmethod
    def __select_or_reject_attr_if_present_key(args, kwargs, modfunc, lookup_attr):
        """
        Processes the select or reject filter only if the attribute exists.

        Parameters:
            args (list): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): The attribute to use in the collection.
            args[3+] (str): Parameters for the Test function.
            modefunc (lambda): Function to use for the collection, used for inversing.
            lookup_attr (bool): Flag for if a particular attribute in the item should be used.

        Returns:
            List of items that either match the test function or do not contain the
            specified attribute.
        """
        seq, func = prepare_select_or_reject(args, kwargs, modfunc, lookup_attr)
        if seq:
            for item in seq:
                if args[2] in item.keys():
                    if func(item):
                        yield item
                else:
                    yield item


    @staticmethod
    def __select_or_reject_key(args, kwargs, inverse):
        """
        Selects or rejects the item only if it contains the key.

        Parameters:
            args (list): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): The attribute to use in the collection.
            args[3+] (str): Parameters for the Test function.
            inverse (lambda): Whether to inverse the result or not.

        Returns:
            List of items that either does or doesn't contain the specified key.
        """
        if args[1]:
            for item in args[1]:
                if inverse:
                    if args[2] not in item.keys():
                        yield item
                else:
                    if args[2] in item.keys():
                        yield item


    @staticmethod
    @contextfilter
    def rejectattr_ifkey(*args, **kwargs):
        """
        Iterates through the collection and only rejects the item if the attribute
        both exists and it matches the supplied test.

        Parameters:
            args (any): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): The attribute to use in the collection.
            args[3+] (any): Parameters for the Test function.

        Returns:
            List of items that do not match the test function or do not contain the
            specified attribute.
        """
        return JinjaFilters.__select_or_reject_attr_if_present_key(
            args,
            kwargs,
            lambda x: not x,
            True
        )


    @staticmethod
    @contextfilter
    def selectattr_ifkey(*args, **kwargs):
        """
        Iterates through the collection and only selects the item if the attribute
        both exists and it matches the supplied test.

        Parameters:
            args (any): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): The attribute to use in the collection.
            args[3+] (any): Parameters for the Test function.

        Returns:
            List of items that match the test function or do not contain the
            specified attribute.
        """
        return JinjaFilters.__select_or_reject_attr_if_present_key(
            args,
            kwargs,
            lambda x: x,
            True
        )


    @staticmethod
    @contextfilter
    def rejectkey(*args, **kwargs):
        """
        Iterates through the collection and rejects the item it it contains the specified key.

        For use if the `selectattr()` function is returning false positives due to its truthiness
        behaviour.

        Parameters:
            args (list): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): Key to select.

        Returns:
            List of items that either does not contain the specified key.
        """
        return JinjaFilters.__select_or_reject_key(
            args,
            kwargs,
            True
        )


    @staticmethod
    @contextfilter
    def selectkey(*args, **kwargs):
        """
        Iterates through the collection and selects the item it it contains the specified key.

        For use if the `selectattr()` function is returning false positives due to its truthiness
        behaviour.

        Parameters:
            args (list): The arguments to process.
            args[0] (any): The context of the Template.
            args[1] (any): The collection.
            args[2] (str): Key to select.

        Returns:
            List of items that contain the specified key.
        """
        return JinjaFilters.__select_or_reject_key(
            args,
            kwargs,
            False
        )


from jinja2 import Environment, Template

from ..extensions.jinja_filters import JinjaFilters
from ..extensions.jinja_tests import JinjaTests
from ..extensions.jinja_functions import JinjaFunctions
from .ruamel_yaml_wrapper import RuamelYAMLWrapper

class JinjaEnvironmentWrapper(Environment):
    """
    Custom implementation for the Jinja2 environment.
    """


    def __init__(self, **kwargs):
        """
        Initialize the environment, and adds the custom Tests and Filters.

        Parameters:
            **kwargs: Additional configuration for the environment.
        """
        super().__init__(**kwargs)

        yaml = RuamelYAMLWrapper()

        self.filters['rejectattr_ifkey'] = JinjaFilters.rejectattr_ifkey
        self.filters['rejectkey'] = JinjaFilters.rejectkey
        self.filters['selectattr_ifkey'] = JinjaFilters.selectattr_ifkey
        self.filters['selectkey'] = JinjaFilters.selectkey

        self.globals.update(contains_gttimes = JinjaTests.contains_gttimes)
        self.globals.update(postindent_spaces = JinjaFilters.postindent_spaces)
        self.globals.update(preindent_spaces = JinjaFilters.preindent_spaces)
        self.globals.update(align_multiline = JinjaFilters.align_multiline)
        self.globals.update(toyaml=yaml.dump_str)

        self.tests['contains'] = JinjaTests.contains
        self.tests['contains_gttimes'] = JinjaTests.contains_gttimes
        self.tests['contains_times'] = JinjaTests.contains_times
        self.tests['endswith'] = JinjaTests.endswith
        self.tests['not_endswith'] = JinjaTests.not_endswith
        self.tests['not_startswith'] = JinjaTests.not_startswith
        self.tests['regex_match'] = JinjaTests.regex_match
        self.tests['startswith'] = JinjaTests.startswith


    def from_file(self, file: str) -> Template:
        """
        Load a Jinja2 template from file.

        Parameters:
            file (str): The file path to load the template from.

        Returns:
            Returns the loaded template.
        """
        with open(file, mode='r') as f:
            return super().from_string(f.read())



import contextlib
import os
import re
from pathlib import Path
from typing import List

from .string_helper import StringHelper


class FileHelper:
    """
    Collection of methods to help with files.
    """


    @staticmethod
    def get_files(path: str, pattern: str, recursive: bool = False):
        """
        Get a file or list of files using a path and search pattern.

        Parameters:
            path (str): The path to search for files in.
            pattern (str): The search pattern to use when searching for
                files within a directory.
            recursive (bool): Flag if the method should search for files in
                sub-folders or just the root folder.

        Returns:
            A list of files for the path.
        """
        if os.path.isfile(path):
            files = [Path(path)]
        else:
            folder = Path(path)

            if recursive:
                files = list(folder.rglob(pattern))
            else:
                files = list(folder.glob(pattern))

        return files


    @staticmethod
    @contextlib.contextmanager
    def open_makedirs(file: str, mode: str):
        """

        """
        # Make sure that the files directory exists.
        os.makedirs(
            FileHelper.get_file_parent(file),
            exist_ok=True
        )

        # Open the file.
        s = open(file, mode)

        # Yield the file stream.
        yield s

        # Close the stream once it is finished.
        s.close()


    @staticmethod
    def get_file_parent(file: str):
        """

        """
        return Path(file).parent


    @staticmethod
    def join_paths(path: str, *paths: list) -> str:
        """

        """
        path = StringHelper.remove_postfix(path, '/')
        path = StringHelper.remove_postfix(path, '\\')

        if not isinstance(paths, list):
            paths = list(paths)

        for index in range(0, len(paths)):
            paths[index] = StringHelper.remove_prefix(paths[index], '/')
            paths[index] = StringHelper.remove_prefix(paths[index], '\\')
            paths[index] = StringHelper.remove_postfix(paths[index], '/')
            paths[index] = StringHelper.remove_postfix(paths[index], '\\')

        return os.path.join(path, *paths)



import os
import re
from typing import Iterator, List, Tuple

from ruamel.yaml.comments import CommentedSeq

from .helpers.string_helper import StringHelper
from .helpers.yaml_comment_helper import YAMLCommentHelper
from .wrappers.ruamel_yaml_wrapper import RuamelYAMLWrapper

# from helpers.string_helper import StringHelper
# from helpers.yaml_comment_helper import YAMLCommentHelper
# from wrappers.ruamel_yaml_wrapper import RuamelYAMLWrapper


class YAMLDocsParser():
    """

    """


    def __init__(self, yaml_parser: RuamelYAMLWrapper,):
        """

        """
        self._yaml_parser = yaml_parser



    def __get_token_from_comment(self, token: any, exclude_comments: list) -> Iterator[Tuple[any, str]]:
        """
        Extracts the comments from the supplied YAML token, and then yields the comment
        back to the caller (alongside the YAML token).

        Parameters:
            token (any): The YAML token to extract the comment for.

        Returns:
            Iterator[Tuple[any, str]]: Value generator that contains the YAML token and the
            extracted comment.
        """
        # Initialize a variable to store the comment block details.
        comment_block_lines = []

        # Iterate through each of the comments.
        for comment in token:
            # Check the comment is not NULL.
            if comment is not None:
                if not isinstance(comment, list):
                    # Format the initial comment.
                    comment_value = YAMLCommentHelper.format_comment(comment.value)
                    comment_value = YAMLCommentHelper.format_comment_lines(comment_value)
                    comment_block_lines.append(comment_value)
                else:
                    for sub_comment in comment:
                        if sub_comment is not None:
                            # Format the initial comment.
                            sub_comment_value = YAMLCommentHelper.format_comment(sub_comment.value)
                            sub_comment_value = YAMLCommentHelper.format_comment_lines(sub_comment_value)
                            comment_block_lines.append(sub_comment_value)

        # Yield the comment
        comment_block_lines = [line for line in comment_block_lines if not line.startswith('---')]
        comment_block = "\n".join(comment_block_lines or list())

        # Return the parsed documentation.
        if exclude_comments:
            if not StringHelper.startswith_multi(comment_block, exclude_comments):
                yield self._yaml_parser.safe_load(comment_block)
        else:
            yield self._yaml_parser.safe_load(comment_block)


    def extract_docs(
        self,
        section: any,
        exclude_comments: list = None,
        parent: any=None,
        property_name: str = 'docs',
    ) -> Iterator[Tuple[any, Tuple[any, str]]]:
        """
        Recursively extracts the YAML comments from the YAML section. For all the YAML properties.

        Parameters:
            section (any): The YAML section to extract all the comments for.
            parent (any): The parent to the current section. Used to determine behaviour.
                If value is set to 'None' then function assumes that it is currently
                executing at the root of the YAML doc.

        Links:
            https://stackoverflow.com/questions/55130078/yaml-parsing-with-comments-for-doc-purpose

        Returns:
            Iterator[Tuple[any, Tuple[any, str]]]: Value generator that contains both the
            supplied section and a Tuple; that contains the corresponding YAML token
            and extracted comment.
        """
        # Extract the comments for a dictionary section.
        if isinstance(section, dict):
            # Loop through all the keys in the dictionary.
            for key in section.keys():
                # Localize the value, done to avoid 'OrderedDict mutated during iteration'.
                val = section[key]

                # Drill down to get the nested comments, for complex values.
                self.extract_docs(val, parent=section, exclude_comments=exclude_comments)

                # Check for a comment in the YAML (iteration) object.
                if isinstance(val, CommentedSeq):
                    # Is there a comment?
                    if val.ca.items and '0' in val.ca.items:
                        for comment in self.__get_token_from_comment(val.ca.items[0], exclude_comments):
                            section[key][property_name] = comment
                    # else:
                    #     section[key][property_name] = None

                # Check if we are currently at the root, or the section is a
                # list of YAML objects (CommentedSeq).
                if parent is None or not isinstance(parent, CommentedSeq):
                    # Re-initialize the 'key: str' pair with a new dict to hold
                    # both the value and docs; 'key: { value: str, docs: {} }'.
                    if isinstance(val, str):
                        section[key] = {
                            'name': key,
                            'value': val
                        }

                # Is there a comments section?
                if section.ca and section.ca.items:
                    # Are there key specific comments?
                    if key in section.ca.items:
                        for comment in self.__get_token_from_comment(section.ca.items[key], exclude_comments):
                            # Check if we are currently at the root, or the section is a
                            # list of YAML objects (CommentedSeq
                            if parent is None or not isinstance(parent, CommentedSeq):
                                # Add the docs to the 'key: str' pair. The pair is re-initialized
                                # above so that the structure is consistent; 'key: { value: str, docs: {} }'.
                                if isinstance(val, str):
                                    section[key][property_name] = comment
                            else:
                                if not isinstance(val, dict):
                                    section[property_name] = comment

        # Extract the comments for a list section.
        elif isinstance(section, list):
            # Iterate through all the items in the YAML list.
            for index, item in enumerate(section):
                # Drill down to get the nested comments, for complex values.
                self.extract_docs(item, parent=section, exclude_comments=exclude_comments)

                # If item is a string re-initialize it with a new dict to hold
                # both the value and the docs.
                if isinstance(item, str):
                    section[index] = {
                        'value': item
                    }

                # TODO: Test if this can be removed? See corresponding section
                # TODO: 'Check for a comment in the YAML (iteration) object.'
                # for comment in self.get_yaml_comments(item):
                #     yield section

                # Are there item specific comments?
                if index in section.ca.items:
                    for comment in self.__get_token_from_comment(section.ca.items[index], exclude_comments):
                        if property_name not in section[index]:
                            section[index][property_name] = comment

        # Get inline comments
        if parent is None and isinstance(section, dict):
            if section.ca.comment is not None:
                for comment in self.__get_token_from_comment(section.ca.comment, exclude_comments):
                    section[property_name] = comment


from setuptools import setup, find_packages

with open('requirements.txt', mode='r') as f:
    requirements = f.read().splitlines()

setup(
    name='embedded-yaml-docs',
    version='0.0.6',
    packages=find_packages(),
    package_dir={
        'embedded-yaml-docs': 'src'
    },
    entry_points='''
        [console_scripts]
        embedded-yaml-docs=src.__main__:main
    ''',
    url='https://github.com/GenesisCoast/embedded-yaml-docs',
    license='MIT',
    install_requires=requirements,
    author='Harry Sanderson',
    author_email='harrysanderson@hotmail.co.uk',
    description='embedded_yaml_docs is a CLI used for generating markdown documentation, from your YAML files. This can be useful for documenting YAML files such as CI pipelines.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        # 'Topic :: Software Development :: Libaries :: Application Frameworks',
        # 'Topic :: Software Development :: Libaries :: Python Modules',
        # 'Topic :: Software Development :: Quality Assurance',
        # 'Topic :: Utilities'
    ]
)



import re


class JinjaTests:
    """
    Collection of custom Jinja Tests.
    """


    @staticmethod
    def regex_match(value: any, pattern: str, flags: int=0) -> bool:
        """
        Tests if the value matches the supplied regex pattern.

        Parameters:
            value (any): The value to test.
            pattern (str): Regex pattern to test against.
            flags (int): Any specific regex flags.

        Returns:
            True if the value matches the regex pattern, else false.
        """
        return re.match(pattern, str(value), flags)


    @staticmethod
    def startswith(value: any, prefix: str) -> bool:
        """
        Tests if the value startswith the supplied prefix.

        Parameters:
            value (str): The value to test.
            prefix (str): Prefix to test against.

        Returns:
            True if the value startswith the supplied prefix.
        """
        return str(value).lower().startswith(prefix)


    @staticmethod
    def endswith(value: any, postfix: str) -> bool:
        """
        Tests if the value endswith the supplied postfix.

        Parameters:
            value (str): The value to test.
            postfix (str): Postfix to test against.

        Returns:
            True if the value endswith the supplied postfix.
        """
        return str(value).lower().endswith(postfix)


    @staticmethod
    def not_startswith(value: any, prefix: str) -> bool:
        """
        Tests if the value does not startswith the supplied prefix.

        Parameters:
            value (str): The value to test.
            prefix (str): Prefix to test against.

        Returns:
            False if the value startswith the supplied prefix.
        """
        return str(value).lower().startswith(prefix)


    @staticmethod
    def not_endswith(value: any, postfix: str) -> bool:
        """
        Tests if the value does not endswith the supplied postfix.

        Parameters:
            value (str): The value to test.
            postfix (str): Postfix to test against.

        Returns:
            False if the value endswith the supplied postfix.
        """
        return str(value).lower().endswith(postfix)


    @staticmethod
    def contains(value: any, substring: str) -> bool:
        """
        Tests if the value contains a substring.

        Parameters:
            value (str): The value to test.
            substring (str): Substring to test against.

        Returns:
            True if the value contains the supplied substring.
        """
        return substring in str(value)


    @staticmethod
    def contains_times(value: any, substring: str, times: int) -> bool:
        """
        Tests if the the value contains the substring a certain amount of times.

        Parameters:
            value (str): The value to test.
            substring (str): Substring to test against.
            times (int): Number of substrings that should be present in the value.

        Returns:
            True if value contains the specified substring a certain amount of times.
        """
        return str(value).count(substring) == int(times)


    @staticmethod
    def contains_gttimes(value: any, substring: str, greater_than: int) -> bool:
        """
        Tests if the number of occurrences of the substring in the value, is greater than
        the supplied number.

        Parameters:
            value (str): The value to test.
            substring (str): Substring to test against.
            greater_than (int): Number of substring occurrences that should be greater than.

        Returns:
            True if the number of substring occurrences is greater than the supplied number.
        """
        return str(value).strip().count(substring) > int(greater_than)